--imports
local assert = require("justo.assert")
local justo = require("justo")
local suite, test, init = justo.suite, justo.test, justo.init
local Parser = require("dogma.syn.Parser")
local SentType = require("dogma.syn.SentType")

--Suite.
return suite("dogma.syn.ExpParser", function()
  local parser, exp

  ----------
  -- init --
  ----------
  init("*", function()
    parser = Parser.new()
  end):title("Create parser")

  -------------
  -- hasId() --
  -------------
  test("hasId()", function(params)
    parser:parse(params[1])
    assert(parser:next():hasId("_")):eq(params[2])
  end):iter(
    {subtitle="# true when id", params={"_", true}},
    {subtitle="# false when id", params={"xyz", false}},
    {subtitle="# true when unary op", params={"+_", true}},
    {subtitle="# false when unary op", params={"+x", false}},
    {subtitle="# true when bin op", params={"x+_", true}},
    {subtitle="# false when bin op", params={"x+y", false}},
    {subtitle="# true when nary op", params={"x(_)", true}},
    {subtitle="# false when nary op", params={"x(y)", false}},
    {subtitle="# true when if op", params={"(if _.key then Key(_.key) end)", true}},
    {subtitle="# false when if op", params={"(if fields.key then Key(fields.key) end)", false}},
    {subtitle="# true when with..do op", params={"(with _.key do\nif x then y\nend)", true}},
    {subtitle="# true when with..do op", params={"(with value do\nif x then y\nelse _.y\nend)", true}},
    {subtitle="# false when with..do op", params={"(with value do\nif x then y\nend)", false}}
  )

  --------------
  -- terminal --
  --------------
  suite("terminal", function()
    suite("(Exp)", function()
      test("when (Exp), ok", function()
        parser:parse("(1+2)")
        exp = parser:next()
        assert(tostring(exp)):eq("(+ 1 2)")
      end)

      test("when (Exp) + ..., ok", function()
        parser:parse("(1+2)+3")
        exp = parser:next()
        assert(tostring(exp)):eq("(+ (+ 1 2) 3)")
      end)

      test("when (), ok", function()
        parser:parse("(\n)")
        exp = parser:next()
        assert(tostring(exp)):eq("()")
      end)

      test("when (exp, exp), ok", function()
        parser:parse("(\n1+2\n3+4\n)")
        exp = parser:next()
        assert(tostring(exp)):eq("((+ 1 2), (+ 3 4))")
      end)

      test("when (exp, exp, error: ) expected", function()
        parser:parse("(\n1+2\n3+4\n")
        assert(function() parser:next() end):raises("anonymous code: ')' expected at the end of code.")
      end)

      test("when (Exp, error: ) expected", function()
        parser:parse("(1+2")
        assert(function() parser:next() end):raises("anonymous code (1, 5): ')' expected.")
      end)
    end)

    suite("if", function()
      test("if Exp then Exp else Exp end", function()
        parser:parse("x + if x == true then a+b else c+d end")
        exp = parser:next()
        assert(tostring(exp)):eq("(+ x (if (== x true) (+ a b) (+ c d)))")
      end)

      test("if Exp then Exp else (exp, exp)", function()
        parser:parse("x + if x == true then a+b else (\nc+d\ne+f\n)")
        exp = parser:next()
        assert(tostring(exp)):eq("(+ x (if (== x true) (+ a b) ((+ c d), (+ e f))))")
      end)

      test("if Exp then\\nExp\\nelse\\nExp\\nend", function()
        parser:parse("x + if x == true then\na+b\nelse\nc+d\nend")
        exp = parser:next()
        assert(tostring(exp)):eq("(+ x (if (== x true) (+ a b) (+ c d)))")
      end)

      test("if Exp then Exp end", function()
        parser:parse("x + if x == true then a+b end")
        exp = parser:next()
        assert(tostring(exp)):eq("(+ x (if (== x true) (+ a b) nil))")
      end)

      test("if Exp then (exp, exp) end", function()
        parser:parse("x + if x == true then (\na+b\nc+d\n)")
        exp = parser:next()
        assert(tostring(exp)):eq("(+ x (if (== x true) ((+ a b), (+ c d)) nil))")
      end)

      test("if Exp else Exp - error - then expected", function()
        parser:parse("x + if a else b")
        assert(function() parser:nextExp() end):raises("anonymous code (1, 10): 'then' expected.")
      end)

      test("if Exp then Exp else Exp - error - end expected", function()
        parser:parse("x + if a then b else c\n\n")
        assert(function() parser:nextExp() end):raises("anonymous code (2, 1): 'end' expected.")
      end)

      test("a + if Exp then Exp else Exp end + e", function()
        parser:parse("a + if b then c else d end + e")
        assert(tostring(parser:nextExp())):eq("(+ (+ a (if b c d)) e)")
      end)
    end)

    suite("with..do", function()
      test("'with' with only ifs", function()
        parser:parse("x + with a do\nif 1 then 2\nif 3 then 4\nend")
        exp = parser:next()
        assert(tostring(exp)):eq("(+ x (with a do 1=>2 3=>4 nil))")
      end)

      test("'with' with ifs and else", function()
        parser:parse("x + with a do\nif 1 then 2\nif 3 then 4\nelse 5end")
        exp = parser:next()
        assert(tostring(exp)):eq("(+ x (with a do 1=>2 3=>4 5))")
      end)
    end):tags("with-do")

    test("self", function()
      parser:parse("self")
      exp = parser:next()
      assert(exp):isTable():has({
        line = 1,
        col = 1,
        type = SentType.EXP
      })
      assert(exp.tree.root):isNotNil()
      assert(tostring(exp)):eq("self")
    end)

    test("x", function()
      parser:parse("x")
      exp = parser:next()
      assert(exp):isTable():has({
        line = 1,
        col = 1,
        type = SentType.EXP
      })
      assert(exp.tree.root):isNotNil()
      assert(tostring(exp)):eq("x")
    end)

    suite("async", function()
      test("fn or proc expected", function()
        parser:parse("x + async pro() end")
        assert(function() exp = parser:next() end):raises("'fn' or 'proc' expected.")
      end)
    end)

    suite("proc", function()
      test("async proc", function()
        parser:parse("x + async proc() end")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x async fn(){})")
      end)

      test("proc", function()
        parser:parse("x + proc() end")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn(){})")
      end)

      test("proc() {}", function()
        parser:parse("x + proc() {}")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn(){})")
      end)

      test("proc() {Exp}", function()
        parser:parse("x + proc() {1+2}")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn(){(+ 1 2)})")
      end)

      test("proc() => Exp", function()
        parser:parse("x + proc() => x(y, z=1)\nend")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn() => (call x y z=1){})")
      end)

      test("proc() {Directive Exp}", function()
        parser:parse("x + proc() {#[!insert /* ignore */] 1}")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):like("table"):like("fn")
      end)
    end)

    suite("fn", function()
      test("async fn", function()
        parser:parse("x + async fn() end")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x async fn(){})")
      end)

      test("fn() => Exp", function()
        parser:parse("x + fn() => x(y)\nend")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn() => (call x y){})")
      end)

      test("fn() => Exp - error, eol skipped", function()
        assert(function()
          parser:parse("x + fn() => x(y) end")
          parser:next()
        end):raises("anonymous code (1, 18): end of line expected.")
      end)

      test("fn() {Exp}", function()
        parser:parse("x + fn() {1+2}")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn(){(+ 1 2)})")
      end)

      test("fn() {}", function()
        parser:parse("x + fn() {}")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn(){})")
      end)

      test("fn() = Exp end", function()
        parser:parse("x + fn() = 1+2 end")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn(){return (+ 1 2)})")
      end)

      test("fn() = (Exp)", function()
        parser:parse("x + fn() = (1+2)")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn(){return (+ 1 2)})")
      end)

      test("fn() {Exp}", function()
        parser:parse("x + fn() {1+2}")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn(){(+ 1 2)})")
      end)

      test("fn() {Directive Exp}", function()
        parser:parse("x + fn() {#[!insert /* ignore */] 1}")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):like("table"):like("fn")
      end)

      test("fn(a,b) end", function()
        parser:parse("x + fn(a,b) end")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn(a, b){})")
      end)

      test("fn(a,b) {}", function()
        parser:parse("x + fn(a,b) {}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn(a, b){})")
      end)

      test("fn()\\nSent\\nend", function()
        parser:parse("x + fn(a, b)\nreturn a+b\nend")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn(a, b){return (+ a b)})")
      end)

      test("fn()\\nSent\\nSent\\n end", function()
        parser:parse("x + fn()\n1\n2\nend")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x fn(){1; 2})")
      end)
    end)

    suite("xml", function()
      test("#!react expected", function()
        parser:parse("<button />")
        assert(function() parser:next() end):raises("anonymous code (1, 1): XML only when #!react specified.")
      end)

      test("<element />", function()
        parser:parse("<button />", nil, {react=true})
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.XML_EXP
        })
        assert(tostring(exp)):eq("(<button></button>)")
      end)

      test("<element /><element />", function()
        parser:parse("<button /><text />", nil, {react=true})
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.XML_EXP
        })
        assert(tostring(exp)):eq("(<button></button>)")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 11,
          type = SentType.XML_EXP
        })
        assert(tostring(exp)):eq("(<text></text>)")
      end)

      test("<element><one/>(children)<two/></element>", function()
        parser:parse("<div><one/>(children)<two/></div>", nil, {react=true})
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.XML_EXP
        })
        assert(tostring(exp)):eq("(<div><one></one>children<two></two></div>)")
      end)

      test("error: < or ( expected into content", function()
        parser:parse("<div><one/>children<two/></div>", nil, {react=true})
        assert(function() parser:next() end):raises("anonymous code (1, 12): '<' or '(' expected.")
      end)

      test("<element keyword=(val1) name=(value2) {prop}=(exp) ...obj ...(exp) />", function()
        parser:parse('<button as="abc" bgColor=("blue") checked fgColor=("red") {prop}=(exp) ...obj ...(exp) />', nil, {react=true})
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.XML_EXP
        })
        assert(tostring(exp)):eq("(<button as=(abc) bgColor=(blue) checked=(checked) fgColor=(red) prop=((. exp prop)) ...obj ...exp></button>)")
      end)

      test("<element></element>", function()
        parser:parse("<button></button>", nil, {react=true})
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.XML_EXP
        })
        assert(tostring(exp)):eq("(<button></button>)")
      end)

      test("<element>content</element>", function()
        parser:parse('<span>"a text"</span>', nil, {react=true})
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.XML_EXP
        })
        assert(tostring(exp)):eq("(<span>a text</span>)")
      end)

      test("<element>\\ncontent\\n</element>", function()
        parser:parse('<span>\n"a text"\n</span>', nil, {react=true})
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.XML_EXP
        })
        assert(tostring(exp)):eq("(<span>a text</span>)")
      end)

      test("<element prop1=(val1) props2=(value2)></element>", function()
        parser:parse('<button bgColor=("blue") checked fgColor=("red")></button>', nil, {react=true})
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.XML_EXP
        })
        assert(tostring(exp)):eq("(<button bgColor=(blue) checked=(checked) fgColor=(red)></button>)")
      end)

      test("<element prop1=(val1) props2=(value2)>content</element>", function()
        parser:parse('<span bgColor=("blue") checked fgColor=("red")>"a text"</span>', nil, {react=true})
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.XML_EXP
        })
        assert(tostring(exp)):eq("(<span bgColor=(blue) checked=(checked) fgColor=(red)>a text</span>)")
      end)

      test("<element></ELEMENT>", function()
        parser:parse("<button></BUTTON>", nil, {react=true})
        assert(function() parser:next() end):raises("anonymous code (1, 11): 'button' expected.")
      end)

      test("<element prop=myvar/>", function()
        parser:parse("<button onClick=myvar />", nil, {react=true})
        assert(tostring(parser:next())):eq("(<button onClick=(myvar)></button>)")
      end)

      test("<element prop=fn/>", function()
        parser:parse("<button onClick=fn() = 123 end />", nil, {react=true})
        assert(tostring(parser:next())):eq("(<button onClick=(fn(){return 123})></button>)")
      end)

      test("<element prop=async/>", function()
        parser:parse("<button onClick=async fn() = 123 end />", nil, {react=true})
        assert(tostring(parser:next())):eq("(<button onClick=(async fn(){return 123})></button>)")
      end)

      test("<element prop={fields}/>", function()
        parser:parse('<button style={color="blue"} />', nil, {react=true})
        assert(tostring(parser:next())):eq("(<button style=({color = blue})></button>)")
      end)

      test("<element prop=[items]/>", function()
        parser:parse('<button style=[1, 2, 3, 6] />', nil, {react=true})
        assert(tostring(parser:next())):eq("(<button style=([1, 2, 3, 6])></button>)")
      end)

      test("<element prop=true/>", function()
        parser:parse('<button active=true />', nil, {react=true})
        assert(tostring(parser:next())):eq("(<button active=(true)></button>)")
      end)

      test("<element prop=false/>", function()
        parser:parse('<button active=false />', nil, {react=true})
        assert(tostring(parser:next())):eq("(<button active=(false)></button>)")
      end)

      test("<element prop=num/>", function()
        parser:parse('<button active=123 />', nil, {react=true})
        assert(tostring(parser:next())):eq("(<button active=(123)></button>)")
      end)

      test("<element prop=text/>", function()
        parser:parse('<button active="bonjour" />', nil, {react=true})
        assert(tostring(parser:next())):eq("(<button active=(bonjour)></button>)")
      end)

      test("<element.subelement prop=text/>", function()
        parser:parse('<Grid.Row prop="value"/>', nil, {react=true})
        assert(tostring(parser:next())):eq("(<Grid.Row prop=(value)></Grid.Row>)")
      end)

      test("<element.subelement prop=text></element.subelement>", function()
        parser:parse('<Grid.Row prop="value"></Grid.Row>', nil, {react=true})
        assert(tostring(parser:next())):eq("(<Grid.Row prop=(value)></Grid.Row>)")
      end)

      test("<element.subelement></element.SUBELEMENT>", function()
        parser:parse("<Grid.Row></GRID.ROW>", nil, {react=true})
        assert(function() parser:next() end):raises("anonymous code (1, 13): 'Grid.Row' expected.")
      end)
    end):tags("xml")

    suite("[for]", function()
      test("[for x in y do z]", function()
        parser:parse("x = [for x in y do z]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(= x [for x in y do z])")
        assert(parser:next()):isNil()
      end)

      test("[\\nfor x in y do z\\n]", function()
        parser:parse("x = [\nfor x in y do z\n]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(= x [for x in y do z])")
        assert(parser:next()):isNil()
      end)

      test("[for a, b in c do d]", function()
        parser:parse("x = [for a, b in c do d]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(= x [for a b in c do d])")
        assert(parser:next()):isNil()
      end)
    end)

    suite("list", function()
      test("x + []", function()
        parser:parse("x + []")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [])")
      end)

      test("x + [\\n]", function()
        parser:parse("x + [\n]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [])")
      end)

      test("x + [123]", function()
        parser:parse("x + [123]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [123])")
      end)

      test("x + [\\n123\\n]", function()
        parser:parse("x + [\n123\n]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [123])")
      end)

      test("x + [123, 456]", function()
        parser:parse("x + [123, 456]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [123, 456])")
      end)

      test("x + [\\n123\\n456\\n]", function()
        parser:parse("x + [\n123\n456\n]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [123, 456])")
      end)

      test("x + [123, 456, 789]", function()
        parser:parse("x + [123, 456, 789]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [123, 456, 789])")
      end)

      test("x + [\\n12, 34\\n56,\\n78]", function()
        parser:parse("x + [\n12, 34\n56,\n78]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [12, 34, 56, 78])")
      end)

      test("x + [123, 456, 789] + y", function()
        parser:parse("x + [123, 456, 789] + y")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ (+ x [123, 456, 789]) y)")
      end)
    end)

    suite("set", function()
      test("x + [||]", function()
        parser:parse("x + [||]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [||])")
      end)

      test("x + [|\\n|]", function()
        parser:parse("x + [|\n|]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [||])")
      end)

      test("x + [|123|]", function()
        parser:parse("x + [|123|]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [|123|])")
      end)

      test("x + [|\\n123\\n|]", function()
        parser:parse("x + [|\n123\n|]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [|123|])")
      end)

      test("x + [|123, 456|]", function()
        parser:parse("x + [|123, 456|]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [|123, 456|])")
      end)

      test("x + [|\\n123\\n456\\n|]", function()
        parser:parse("x + [|\n123\n456\n|]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [|123, 456|])")
      end)

      test("x + |123, 456, 789|", function()
        parser:parse("x + [|123, 456, 789|]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [|123, 456, 789|])")
      end)

      test("x + [|\\n12, 34\\n56,\\n78|]", function()
        parser:parse("x + [|\n12, 34\n56,\n78|]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x [|12, 34, 56, 78|])")
      end)

      test("x + [|123, 456, 789|] + y", function()
        parser:parse("x + [|123, 456, 789|] + y")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ (+ x [|123, 456, 789|]) y)")
      end)
    end)

    suite("map", function()
      test("{}", function()
        parser:parse("x + {}")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()

        assert(tostring(exp)):eq("(+ x {})")
        assert(parser:next()):isNil()
      end)

      test("{\\n}", function()
        parser:parse("x + {\n}")

        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()

        assert(tostring(exp)):eq("(+ x {})")
        assert(parser:next()):isNil()
      end)

      test("{name}", function()
        parser:parse("x + {variable}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {variable = variable})")
      end)

      test("{if cond then name = value}", function()
        parser:parse("x + {if x then field = value}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {[field = value]})")
      end)

      test("{...variable}", function()
        parser:parse("x + {...variable}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {(... variable)})")
      end)

      test("{{name} = obj}", function()
        parser:parse("x + {{fld} = obj}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {fld = (. obj fld)})")
        assert(parser:next()):isNil()
      end)

      test("{{keyword} = obj}", function()
        parser:parse("x + {{then} = obj}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {then = (. obj then)})")
      end)

      test("{{{name}}}", function()
        parser:parse("x + {{{fld}}}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {fld = (. fld fld)})")
        assert(parser:next()):isNil()
      end)

      test("{[field] = exp}", function()
        parser:parse("x + {[x+y] = a+b}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {[(+ x y)] = (+ a b)})")
        assert(parser:next()):isNil()
      end)

      test("{[field] =\\n  map}", function()
        parser:parse("x + {\n  abc =\n    [x+y] =\n      z = 123\n}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {abc = {[(+ x y)] = {z = 123}}})")
        assert(parser:next()):isNil()
      end)

      test("x + {one = 1}", function()
        parser:parse("x + {one = 1}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {one = 1})")
      end)

      test("x + {\\none = 1\\n}", function()
        parser:parse("x + {\none = 1\n}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {one = 1})")
      end)

      test("x + {name = 1, keyword = 2}", function()
        parser:parse("x + {name = 1, then = 2}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {name = 1, then = 2})")
      end)

      test("x + {\\none = 1\\ntwo = 2\\n}", function()
        parser:parse("x + {\none = 1\ntwo = 2\n}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {one = 1, two = 2})")
      end)

      test("x + {\\none = 1\\n\\ntwo = 2\\n}", function()
        parser:parse("x + {\none = 1\n\ntwo = 2\n}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {one = 1, two = 2})")
      end)

      test("x + {one = 1, two = 2\\nthree = 3,\\nfour = 4}", function()
        parser:parse("x + {one = 1, two = 2\nthree = 3,\nfour=4}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {one = 1, two = 2, three = 3, four = 4})")
      end)

      test("x + {\\none = 1\\ntwo = 2\\nthree = 3\\n}", function()
        parser:parse("x + {\none = 1\ntwo = 2\nthree = 3\n}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {one = 1, two = 2, three = 3})")
      end)

      test("map with fields without {}", function()
        parser:parse("x + {\na =\n  one = 1\n  two = 2\n  {three} = 3\nb = 4}")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(exp.tree.root):isNotNil()
        assert(tostring(exp)):eq("(+ x {a = {one = 1, two = 2, three = (. 3 three)}, b = 4})")
        assert(parser:next()):isNil()
      end)
    end)
  end)

  --------------
  -- unary op --
  --------------
  suite("unary op", function()
    test("...", function()
      parser:parse("...args")
      assert(tostring(parser:next())):eq("(... args)")
    end)

    test("<<< list", function()
      parser:parse("<<< list")
      assert(tostring(parser:next())):eq("(<<< list)")
    end)

    test(">>> list", function()
      parser:parse(">>> list")
      assert(tostring(parser:next())):eq("(>>> list)")
    end)

    suite("+", function()
      test("+x", function()
        parser:parse("+x")
        assert(tostring(parser:next())):eq("(+ x)")
      end)

      test("x + +y", function()
        parser:parse("x + +y")
        assert(tostring(parser:next())):eq("(+ x (+ y))")
      end)

      test("x + +y + z", function()
        parser:parse("x + +y + z")
        assert(tostring(parser:next())):eq("(+ (+ x (+ y)) z)")
      end)
    end)

    suite("-", function()
      test("-x", function()
        parser:parse("-x")
        assert(tostring(parser:next())):eq("(- x)")
      end)

      test("x + -y", function()
        parser:parse("x + -y")
        assert(tostring(parser:next())):eq("(+ x (- y))")
      end)

      test("x + -y + z", function()
        parser:parse("x + -y + z")
        assert(tostring(parser:next())):eq("(+ (+ x (- y)) z)")
      end)
    end)

    test("not x", function()
      parser:parse("not x")
      assert(tostring(parser:next())):eq("(not x)")
    end)

    test(". Name", function()
      parser:parse(".x")
      assert(tostring(parser:nextExp())):eq("(. x)")
    end)

    test(": Name", function()
      parser:parse(":x")
      assert(tostring(parser:nextExp())):eq("(: x)")
    end)

    test("! Name", function()
      parser:parse("!x")
      assert(tostring(parser:nextExp())):eq("(! x)")
    end)

    test("$ Name", function()
      parser:parse("$x")
      assert(tostring(parser:nextExp())):eq("($ x)")
    end)

    test(":123 - error", function()
      parser:parse(":123")
      assert(function()
        parser:nextExp()
      end):raises("anonymous code (1, 2): '.', ':', '!' and '$' must be followed by identifier.")
    end)
  end):tags("unaryop")

  ------------
  -- bin op --
  ------------
  suite("bin op", function()
    test("= x + y must raise error", function()
      parser:parse("= x + y")
      assert(function() parser:next() end):raises("the = operator can't be the first node in the expression.")
    end)

    test("??", function()
      parser:parse("x ?? y ?? z")
      assert(tostring(parser:next())):eq("(?? (?? x y) z)")
    end)

    test("list <<< item", function()
      parser:parse("list <<< item")
      assert(tostring(parser:next())):eq("(<<< list item)")
    end)

    test("stack >>> item", function()
      parser:parse("item >>> list")
      assert(tostring(parser:next())):eq("(>>> item list)")
    end)

    test("x == 123", function()
      parser:parse("x == 123")
      assert(tostring(parser:next())):eq("(== x 123)")
    end)

    test("x ==~ 123", function()
      parser:parse("x ==~ 123")
      assert(tostring(parser:next())):eq("(==~ x 123)")
    end)

    test("x = 123", function()
      parser:parse("x = 123")
      assert(tostring(parser:next())):eq("(= x 123)")
    end)

    test("x =~ ITEM", function()
      parser:parse("x =~ ITEM")
      assert(tostring(parser:next())):eq("(=~ x ITEM)")
    end)

    test("x ?= 123", function()
      parser:parse("x ?= 123")
      assert(tostring(parser:next())):eq("(?= x 123)")
    end)

    test(":x .= 123", function()
      parser:parse(":x .= 123")
      assert(tostring(parser:next())):eq("(.= (: x) 123)")
    end)

    test("x := 123", function()
      parser:parse("x := 123")
      assert(tostring(parser:next())):eq("(:= x 123)")
    end)

    test("x += 123", function()
      parser:parse("x += 123")
      assert(tostring(parser:next())):eq("(+= x 123)")
    end)

    test("?", function()
      parser:parse("x?y")
      assert(tostring(parser:next())):eq("(? x y)")
    end)

    suite(".", function()
      test("x.y", function()
        parser:parse("x.y")
        assert(tostring(parser:next())):eq("(. x y)")
      end)

      test("(x+y).z", function()
        parser:parse("(x+y).z")
        assert(tostring(parser:next())):eq("(. (+ x y) z)")
      end)

      test("a+(b+c).d", function()
        parser:parse("a+(b+c).d")
        assert(tostring(parser:next())):eq("(+ a (. (+ b c) d))")
      end)
    end)

    suite(":", function()
      test("x:y", function()
        parser:parse("x:y")
        assert(tostring(parser:next())):eq("(: x y)")
      end)

      test("(x+y):z", function()
        parser:parse("(x+y):z")
        assert(tostring(parser:next())):eq("(: (+ x y) z)")
      end)

      test("a+(b+c):d", function()
        parser:parse("a+(b+c):d")
        assert(tostring(parser:next())):eq("(+ a (: (+ b c) d))")
      end)
    end)

    suite("!", function()
      test("x!y", function()
        parser:parse("x!y")
        assert(tostring(parser:next())):eq("(! x y)")
      end)

      test("(x+y)!z", function()
        parser:parse("(x+y)!z")
        assert(tostring(parser:next())):eq("(! (+ x y) z)")
      end)

      test("a+(b+c)!d", function()
        parser:parse("a+(b+c)!d")
        assert(tostring(parser:next())):eq("(+ a (! (+ b c) d))")
      end)
    end)

    test("x + y", function()
      parser:parse("x+y")
      assert(tostring(parser:next())):eq("(+ x y)")
    end)

    test("1+2*3 - different precedence", function()
      parser:parse("1 + 2 * 3")
      exp = parser:nextExp()
      assert(tostring(exp)):eq("(+ 1 (* 2 3))")
    end)

    test("1+2*3 - different precedence", function()
      parser:parse("1 + 2 * 3")
      exp = parser:nextExp()
      assert(tostring(exp)):eq("(+ 1 (* 2 3))")
    end)

    test("1*2+3 - different precedence", function()
      parser:parse("1 * 2 + 3")
      exp = parser:nextExp()
      assert(tostring(exp)):eq("(+ (* 1 2) 3)")
    end)

    test("1+2-3 - same precedence and left assoc", function()
      parser:parse("1 + 2 - 3")
      exp = parser:nextExp()
      assert(tostring(exp)):eq("(- (+ 1 2) 3)")
    end)

    test("a in b", function()
      parser:parse("a in b")
      assert(tostring(parser:nextExp())):eq("(in a b)")
    end)

    test("a not in b", function()
      parser:parse("a not in b")
      assert(tostring(parser:nextExp())):eq("(notin a b)")
    end)

    test("a like b", function()
      parser:parse("a like b")
      assert(tostring(parser:nextExp())):eq("(like a b)")
    end)

    test("a like `pattern`", function()
      parser:parse("a like `pattern`")
      assert(tostring(parser:nextExp())):eq("(like a pattern)")
    end)

    test("a not like b", function()
      parser:parse("a not like b")
      assert(tostring(parser:nextExp())):eq("(notlike a b)")
    end)

    suite("is", function()
      test("a is b", function()
        parser:parse("a is b")
        assert(tostring(parser:nextExp())):eq("(is a b)")
      end)

      test("a is [b, c]", function()
        parser:parse("a is [b, c]")
        assert(tostring(parser:nextExp())):eq("(is a [b, c])")
      end)

      test("a is not b", function()
        parser:parse("a is not b")
        assert(tostring(parser:nextExp())):eq("(isnot a b)")
      end)
    end)

    suite("indexing", function()
      test("a?[b]", function()
        parser:parse("a?[b]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(tostring(exp)):eq("(?[] a b)")
      end)

      test("a[b]", function()
        parser:parse("a[b]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(tostring(exp)):eq("([] a b)")
      end)

      test("a[b + c]", function()
        parser:parse("a[b + c]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(tostring(exp)):eq("([] a (+ b c))")
      end)

      test("a + b[c]", function()
        parser:parse("a + b[c]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(tostring(exp)):eq("(+ a ([] b c))")
      end)

      test("a + b[c + d]", function()
        parser:parse("a + b[c + d]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(tostring(exp)):eq("(+ a ([] b (+ c d)))")
      end)

      test("a[b, c]", function()
        parser:parse("a[b, c]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(tostring(exp)):eq("([] a b c)")
      end)

      test("a[b + c, d]", function()
        parser:parse("a[b + c, d]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(tostring(exp)):eq("([] a (+ b c) d)")
      end)

      test("a + b[c, d]", function()
        parser:parse("a + b[c, d]")
        exp = parser:next()
        assert(exp):isTable():has({
          line = 1,
          col = 1,
          type = SentType.EXP
        })
        assert(tostring(exp)):eq("(+ a ([] b c d))")
      end)
    end)
  end)

  ----------
  -- misc --
  ----------
  suite("misc", function()
    test("not .value.includes(val)", function()
      parser:parse("not .value.includes(val)")
      assert(tostring(parser:next())):eq("(not (call (. (. value) includes) val))")
    end)

    test("x and y", function()
      parser:parse("x and y")
      assert(tostring(parser:next())):eq("(and x y)")
    end)

    test("x && y", function()
      parser:parse("x+y && a+b")
      assert(tostring(parser:next())):eq("(&& (+ x y) (+ a b))")
    end)

    test("x and not y", function()
      parser:parse("x and not y")
      assert(tostring(parser:next())):eq("(and x (not y))")
    end)
  end)

  test("a=b=c - same precedence and right assoc", function()
    parser:parse("a = b = c")
    exp = parser:nextExp()
    assert(tostring(exp)):eq("(= a (= b c))")
  end)

  suite("(Exp)", function()
    test("(1 + 2 * 3)", function()
      parser:parse("(1 + 2 * 3)")
      exp = parser:nextExp()
      assert(tostring(exp)):eq("(+ 1 (* 2 3))")
    end)

    test("(1 + 2 * 3 - error - missing last )", function()
      parser:parse("(1 + 2 * 3")
      assert(function()
        parser:nextExp()
      end):raises("anonymous code (1, 11): ')' expected.")
    end)

    test("(1 + 2 * 3 a - error", function()
      parser:parse("(1 + 2 * 3 a")
      assert(function()
        parser:nextExp()
      end):raises("anonymous code (1, 12): invalid terminal node for well-formed expression.")
    end)
  end)

  test("1 + 2\\n*3", function()
    parser:parse("1 + 2\n* 3")
    assert(tostring(parser:nextExp())):eq("(+ 1 2)")
  end)

  test("1 + 2 -\\n3", function()
    parser:parse("1 + 2 -\n3")
    assert(tostring(parser:nextExp())):eq("(- (+ 1 2) 3)")
  end)

  test("1 + 2 -\\n - error", function()
    parser:parse("1 + 2 -\n")
    assert(function()
      parser:nextExp()
    end):raises("anonymous code (1, 1): incomplete expression.")
  end)

  ----------
  -- call --
  ----------
  suite("call", function()
    test("a()", function()
      parser:parse("a()")
      assert(tostring(parser:nextExp())):eq("(call a)")
    end)

    test(".a()", function()
      parser:parse(".a()")
      assert(tostring(parser:next())):eq("(call (. a))")
    end)

    test("a(b)", function()
      parser:parse("a(b)")
      assert(tostring(parser:nextExp())):eq("(call a b)")
    end)

    test("a(b,c)", function()
      parser:parse("a(b, c)")
      assert(tostring(parser:nextExp())):eq("(call a b c)")
    end)

    test("a(b=1,c=2)", function()
      parser:parse("a(b=1, c=2)")
      assert(tostring(parser:nextExp())):eq("(call a b=1 c=2)")
    end)

    test("a(b,c,d=1,e=2)", function()
      parser:parse("a(b, c, d=1, e=2)")
      assert(tostring(parser:nextExp())):eq("(call a b c d=1 e=2)")
    end)

    test("a(b=,c)", function()
      parser:parse("a(b=, c)")
      assert(tostring(parser:nextExp())):eq("(call a b=b c=c)")
    end)

    test("a(b=1,c)", function()
      parser:parse("a(b=1, c)")
      assert(tostring(parser:nextExp())):eq("(call a b=1 c=c)")
    end)

    test("a(\\nb\\n)", function()
      parser:parse("a(\nb\n)")
      assert(tostring(parser:nextExp())):eq("(call a b)")
    end)

    test("a(\\nb)", function()
      parser:parse("a(\nb)")
      assert(tostring(parser:nextExp())):eq("(call a b)")
    end)

    test("a(\\nb\\nc\\n)", function()
      parser:parse("a(\nb\nc\n)")
      assert(tostring(parser:nextExp())):eq("(call a b c)")
    end)

    test("a(\\nb,\\nc\\n)", function()
      parser:parse("a(\nb,\nc\n)")
      assert(tostring(parser:nextExp())):eq("(call a b c)")
    end)

    test("a(\\nb\\nc)", function()
      parser:parse("a(\nb\nc)")
      assert(tostring(parser:nextExp())):eq("(call a b c)")
    end)

    test("a+b()", function()
      parser:parse("a + b()")
      assert(tostring(parser:nextExp())):eq("(+ a (call b))")
    end)

    test("a+b(c)", function()
      parser:parse("a + b(c)")
      assert(tostring(parser:nextExp())):eq("(+ a (call b c))")
    end)

    test("a+b()+c", function()
      parser:parse("a + b() + c")
      assert(tostring(parser:nextExp())):eq("(+ (+ a (call b)) c)")
    end)

    test("a+b()*c", function()
      parser:parse("a + b() * c")
      assert(tostring(parser:nextExp())):eq("(+ a (* (call b) c))")
    end)

    test("a+b()*c+d", function()
      parser:parse("a + b() * c + d")
      assert(tostring(parser:nextExp())):eq("(+ (+ a (* (call b) c)) d)")
    end)
  end)

  --------------
  -- native() --
  --------------
  suite("native()", function()
    test("native(Literal)", function()
      parser:parse('native("new Proxy(target, handler)")')
      assert(tostring(parser:next())):eq('(native "new Proxy(target, handler)")')
    end)

    test("native() - error", function()
      parser:parse("native()")
      assert(function() parser:next() end):raises("anonymous code (1, 8): literal expected.")
    end)
  end):tags("native")

  -------------
  -- const() --
  -------------
  suite("const()", function()
    test("const(Exp)", function()
      parser:parse("x = const(obj)")
      assert(tostring(parser:next())):eq("(= x (const obj))")
      assert(parser:next()):isNil()
    end)

    test("const(\\nExp\\n)", function()
      parser:parse("x = const(\nobj\n)")
      assert(tostring(parser:next())):eq("(= x (const obj))")
      assert(parser:next()):isNil()
    end)
  end):tags("const")

  ------------
  -- copy() --
  ------------
  suite("copy()", function()
    test("copy(Exp)", function()
      parser:parse("x = copy(obj)")
      assert(tostring(parser:next())):eq("(= x (copy obj))")
      assert(parser:next()):isNil()
    end)

    test("copy(\\nExp\\n)", function()
      parser:parse("x = copy(\nobj\n)")
      assert(tostring(parser:next())):eq("(= x (copy obj))")
      assert(parser:next()):isNil()
    end)
  end):tags("copy")

  -----------
  -- new() --
  -----------
  suite("new()", function()
    test("new(Exp)", function()
      parser:parse("new(1+2)")
      assert(tostring(parser:next())):eq("(new (+ 1 2))")
    end)

    test("new(\\nExp\\n)", function()
      parser:parse("new(\n1+2\n)")
      assert(tostring(parser:next())):eq("(new (+ 1 2))")
    end)
  end):tags("new")

  ----------------
  -- overload() --
  ----------------
  suite("overload()", function()
    test("overload()", function()
      parser:parse("overload()")
      assert(tostring(parser:next())):eq("(overload)")
    end)

    test("overload(name)", function()
      parser:parse("overload(name)")
      assert(tostring(parser:next())):eq("(overload name)")
    end)

    test("overload(?name)", function()
      parser:parse("overload(?name)")
      assert(tostring(parser:next())):eq("(overload ?name)")
    end)

    test("overload(...)", function()
      parser:parse("overload(...)")
      assert(tostring(parser:next())):eq("(overload ...)")
    end)

    test("overload(name, ?name, ...)", function()
      parser:parse("overload(name, ?name, ...)")
      assert(tostring(parser:next())):eq("(overload name ?name ...)")
    end)
  end):tags("overload")

  -------------
  -- await() --
  -------------
  suite("await()", function()
    test("await() - with no argument", function()
      parser:parse("await()")
      assert(function() parser:next() end):raises("anonymous code (1, 7): 1 argument(s) expected. Received: 0.")
    end)

    test("await(Exp)", function()
      parser:parse("await(1+2)")
      assert(tostring(parser:next())):eq("(await (+ 1 2))")
    end)

    test("await(\\nExp\\n)", function()
      parser:parse("await(\n1+2\n)")
      assert(tostring(parser:next())):eq("(await (+ 1 2))")
    end)

    test("await(Exp, Exp)", function()
      parser:parse("await(1+2, 3+4)")
      assert(tostring(parser:next())):eq("(mawait (+ 1 2) (+ 3 4))")
    end)

    test("await(...Exp)", function()
      parser:parse("await(...aList)")
      assert(tostring(parser:next())):eq("(mawait aList)")
    end)
  end):tags("await")

  --------------
  -- pawait() --
  --------------
  suite("pawait()", function()
    test("pawait() - with no argument", function()
      parser:parse("pawait()")
      assert(function() parser:next() end):raises("anonymous code (1, 8): 1 argument(s) expected. Received: 0.")
    end)

    test("pawait(Exp)", function()
      parser:parse("pawait(1+2)")
      assert(tostring(parser:next())):eq("(pawait (+ 1 2))")
    end)

    test("pawait(\\nExp\\n)", function()
      parser:parse("pawait(\n1+2\n)")
      assert(tostring(parser:next())):eq("(pawait (+ 1 2))")
    end)

    test("pawait(Exp, Exp)", function()
      parser:parse("pawait(1+2, 3+4)")
      assert(tostring(parser:next())):eq("(mpawait (+ 1 2) (+ 3 4))")
    end)

    test("pawait(...Exp)", function()
      parser:parse("pawait(...aList)")
      assert(tostring(parser:next())):eq("(mpawait aList)")
    end)
  end):tags("pawait")

  -------------
  -- cwait() --
  -------------
  suite("cwait()", function()
    test("cwait(Exp)", function()
      parser:parse("cwait(1+2)")
      assert(tostring(parser:next())):eq("(cwait (+ 1 2))")
    end)

    test("cwait(\\nExp\\n)", function()
      parser:parse("cwait(\n1+2\n)")
      assert(tostring(parser:next())):eq("(cwait (+ 1 2))")
    end)
  end):tags("cwait")

  --------------
  -- pcwait() --
  --------------
  suite("pcwait()", function()
    test("pcwait(Exp)", function()
      parser:parse("pcwait(1+2)")
      assert(tostring(parser:next())):eq("(pcwait (+ 1 2))")
    end)

    test("pcwait(\\nExp\\n)", function()
      parser:parse("pcwait(\n1+2\n)")
      assert(tostring(parser:next())):eq("(pcwait (+ 1 2))")
    end)
  end):tags("pcwait")

  -------------
  -- pwait() --
  -------------
  test("pwait(Exp)", function()
    parser:parse("pwait(1+2)")
    assert(tostring(parser:next())):eq("(pwait (+ 1 2))")
  end):tags("pwait")

  -------------
  -- use() --
  -------------
  test("use(Exp)", function()
    parser:parse("use(path)")
    assert(tostring(parser:next())):eq("(use path)")
  end):tags("use")

  ------------------
  -- self.defer() --
  ------------------
  test("self.defer(Exp)", function()
    parser:parse("self.defer(1+2)")
    assert(tostring(parser:next())):eq("(self.defer (+ 1 2))")
  end)

  ---------------------
  -- self.deferred() --
  ---------------------
  test("self.deferred()", function()
    parser:parse("self.deferred()")
    assert(tostring(parser:next())):eq("(self.deferred)")
  end)

  -------------
  -- peval() --
  -------------
  suite("peval()", function()
    test("peval(Exp)", function()
      parser:parse("peval(1+2)")
      assert(tostring(parser:next())):eq("(peval (+ 1 2))")
    end)

    test("peval() - error", function()
      parser:parse("peval()")
      assert(function() parser:next() end):raises("anonymous code (1, 7): invalid expression node.")
    end)

    test("peval(1+2 - error", function()
      parser:parse("peval(1+2")
      assert(function() parser:next() end):raises("anonymous code (1, 10): ')' expected.")
    end)
  end)

  -------------------
  -- json.decode() --
  -------------------
  suite("json.decode()", function()
    test("json.decode(Exp)", function()
      parser:parse('json.decode("{}")')
      assert(tostring(parser:next())):eq('(call (. json decode) {})')
    end)

    test("json.decode<MyStruct>(Exp)", function()
      parser:parse('json.decode<MyStruct>("{}")')
      assert(tostring(parser:next())):eq('(json.decode MyStruct {})')
    end)
  end)

  -------------
  -- catch() --
  -------------
  suite("catch()", function()
    test("catch(Exp)", function()
      parser:parse("catch(1+2)")
      assert(tostring(parser:next())):eq("(catch (+ 1 2))")
    end)

    test("catch() - error", function()
      parser:parse("catch()")
      assert(function() parser:next() end):raises("anonymous code (1, 7): invalid expression node.")
    end)
  end)

  ------------
  -- cast() --
  ------------
  suite("cast()", function()
    test("cast<Type>(value)", function()
      parser:parse("cast<Type>(x)")
      assert(tostring(parser:next())):eq("(cast Type x)")
    end)
  end):tags("cast")

  ------------
  -- emit() --
  ------------
  suite("emit()", function()
    test("emit(Exp, Exp)", function()
      parser:parse('emit("evento", 123)')
      assert(tostring(parser:next())):eq("(emit evento 123)")
    end)

    test("emit(Exp, literal map)", function()
      parser:parse('emit("evento", x=, y=1, z)')
      assert(tostring(parser:next())):eq("(emit evento {x=x, y=1, z=z})")
    end)
  end):tags("emit")

  --------------
  -- expect() --
  --------------
  suite("expect()", function()
    test("expect(Exp)", function()
      parser:parse("expect(x)")
      assert(tostring(parser:next())):eq("(expect x nil)")
    end)

    test("expect(Exp, Exp)", function()
      parser:parse("expect(x, num)")
      assert(tostring(parser:next())):eq("(expect x num)")
    end)
  end)

  -------------
  -- throw() --
  -------------
  suite("throw()", function()
    test("throw(Exp)", function()
      parser:parse("throw(1+2)")
      assert(tostring(parser:next())):eq("(throw (+ 1 2))")
    end)

    test("throw(Exp, Exp)", function()
      parser:parse("throw(1+2, 3+4)")
      assert(tostring(parser:next())):eq("(throw (+ 1 2) (+ 3 4))")
    end)

    test("throw(Exp, Exp, Exp)", function()
      parser:parse("throw(1+2, 3+4, 5+6)")
      assert(tostring(parser:next())):eq("(throw (+ 1 2) (+ 3 4) (+ 5 6))")
    end)
  end):tags("throw")

  ----------
  -- pack --
  ----------
  suite("pack", function()
    suite("list", function()
      test("exp[*]", function()
        parser:parse("obj[*]")
        assert(tostring(parser:next())):eq("(packl obj)")
      end)

      test("exp[*,item]", function()
        parser:parse("obj[*,item]")
        assert(tostring(parser:next())):eq("(packl obj item)")
      end)

      test("exp[*,item,item]", function()
        parser:parse("obj[*,item1,item2]")
        assert(tostring(parser:next())):eq("(packl obj item1 item2)")
      end)
    end)

    suite("map", function()
      test("exp{}", function()
        parser:parse("obj{}")
        assert(tostring(parser:next())):eq("(pack obj)")
      end)

      test("exp{*}", function()
        parser:parse("obj{*}")
        assert(tostring(parser:next())):eq("(pack obj .*)")
      end)

      test("exp{\\n*\\n}", function()
        parser:parse("obj{\n*\n}")
        assert(tostring(parser:next())):eq("(pack obj .*)")
      end)

      test("exp{*field}", function()
        parser:parse("obj{*field}")
        assert(function() parser:next() end):raises("anonymous code (1, 6): ',', end of line or '}' expected.")
      end)

      test("exp{*,name}", function()
        parser:parse("obj{*,field}")
        assert(tostring(parser:next())):eq("(pack obj .* .field=field)")
      end)

      test("exp{\\n*\\nname\\n}", function()
        parser:parse("obj{\n*\nfield\n}")
        assert(tostring(parser:next())):eq("(pack obj .* .field=field)")
      end)

      test("exp{\\n*\\nname}", function()
        parser:parse("obj{\n*\nfield}")
        assert(tostring(parser:next())):eq("(pack obj .* .field=field)")
      end)

      test("exp{\\n*,\\nname\\n}", function()
        parser:parse("obj{\n*,\nfield\n}")
        assert(tostring(parser:next())):eq("(pack obj .* .field=field)")
      end)

      test("exp{*,{name}=exp}", function()
        parser:parse("obj{*,{field}=exp}")
        assert(tostring(parser:next())):eq("(pack obj .* .field=(. exp field))")
      end)

      test("exp{*,name=value,keyword=value}", function()
        parser:parse("obj{*,name=12,then=34}")
        assert(tostring(parser:next())):eq("(pack obj .* .name=12 .then=34)")
      end)

      test("exp{*,name=value,keyword=value,-field,...obj}", function()
        parser:parse("obj{*,name=12,then=34,-toRemove, ...obj}")
        assert(tostring(parser:next())):eq("(pack obj .* .name=12 .then=34 -.toRemove (pack obj))")
      end)

      test("exp{**,name=value,keyword=value,-field,...obj}", function()
        parser:parse("obj{**,name=12,then=34,-toRemove, ...obj}")
        assert(tostring(parser:next())):eq("(pack obj .** .name=12 .then=34 -.toRemove (pack obj))")
      end)

      test("exp{name}", function()
        parser:parse("obj{name}")
        assert(tostring(parser:next())):eq("(pack obj .name)")
      end)

      test("exp{name,name=value,...obj}", function()
        parser:parse("obj{name,name2=123,...obj}")
        assert(tostring(parser:next())):eq("(pack obj .name .name2=123 (pack obj))")
      end)

      test("exp{.name}", function()
        parser:parse("obj{.name}")
        assert(tostring(parser:next())):eq("(pack obj .name)")
      end)

      test("exp{:name}", function()
        parser:parse("obj{:name}")
        assert(tostring(parser:next())):eq("(pack obj :name)")
      end)

      test("exp{name,.name,:name}", function()
        parser:parse("obj{one,.two,:three}")
        assert(tostring(parser:next())):eq("(pack obj .one .two :three)")
      end)
    end)
  end):tags("pack")

  ------------
  -- update --
  ------------
  suite("update", function()
    suite("error", function()
      test("obj.{name:=value} - error", function()
        parser:parse("obj.{name:=1+2}")
        assert(function() parser:next() end):raises("anonymous code (1, 10): ',' or '}' expected.")
      end)

      test(".{name}", function()
        parser:parse(".{name}")
        assert(function() parser:next() end):raises("anonymous code (1, 1): '.{', ':{' and '!{' are not unary ops.")
      end)

      test("obj.{}", function()
        parser:parse("obj.{}")
        assert(function() parser:next() end):raises("anonymous code (1, 6): id expected.")
      end)

      test("obj.{\\n}", function()
        parser:parse("obj.{\n}")
        assert(function() parser:next() end):raises("anonymous code (2, 1): id expected.")
      end)
    end)

    test("obj.{name}", function()
      parser:parse("obj.{name}")
      assert(tostring(parser:next())):eq("(update obj .name)")
    end)

    test("obj:{name}", function()
      parser:parse("obj:{name}")
      assert(tostring(parser:next())):eq("(update obj :name)")
    end)

    test("obj!{name}", function()
      parser:parse("obj!{name}")
      assert(tostring(parser:next())):eq("(update obj !name)")
    end)

    test("obj.{name=value}", function()
      parser:parse("obj.{name=1+2}")
      assert(tostring(parser:next())):eq("(update obj .name=(+ 1 2))")
    end)

    test("obj.{name::=value}", function()
      parser:parse("obj.{name::=1+2}")
      assert(tostring(parser:next())):eq("(update obj .name::=(+ 1 2))")
    end)

    test("obj.{name.=value}", function()
      parser:parse("obj.{name.=1+2}")
      assert(tostring(parser:next())):eq("(update obj .name.=(+ 1 2))")
    end)

    test("obj.{name?=value}", function()
      parser:parse("obj.{name?=1+2}")
      assert(tostring(parser:next())):eq("(update obj .name?=(+ 1 2))")
    end)

    test("obj.{{name}=value}", function()
      parser:parse("obj.{{name}=1+2}")
      assert(tostring(parser:next())):eq("(update obj {.name}=(+ 1 2))")
    end)

    test("obj.{{name1, name2}=value}", function()
      parser:parse("obj.{{name1, name2}=1+2}")
      assert(tostring(parser:next())):eq("(update obj {.name1}=(+ 1 2) {.name2}=(+ 1 2))")
    end)

    test("obj.{(name)=value}", function()
      parser:parse("obj.{(name)=1+2}")
      assert(tostring(parser:next())):eq("(update obj (.name)=(+ 1 2))")
    end)

    test("obj.{(name1, name2)=value}", function()
      parser:parse("obj.{(name1, name2)=1+2}")
      assert(tostring(parser:next())):eq("(update obj (.name1)=(+ 1 2) (.name2)=(+ 1 2))")
    end)

    test("obj.{one, two=1+2, three::=3+4, four?=5+6}", function()
      parser:parse("obj.{one, two=1+2, three::=3+4, four?=5+6}")
      assert(tostring(parser:next())):eq("(update obj .one .two=(+ 1 2) .three::=(+ 3 4) .four?=(+ 5 6))")
    end)

    test("obj.{one;two}", function()
      parser:parse("obj.{one;two}")
      assert(function() parser:next() end):raises("anonymous code (1, 9): ',' or '}' expected.")
    end)

    test("obj.{\\none\\ntwo}", function()
      parser:parse("obj.{\none\ntwo}")
      assert(tostring(parser:next())):eq("(update obj .one .two)")
    end)
  end):tags("update")
end):tags("exp")
